#!/usr/bin/env python
# -*- coding: latin-1 -*-

[{'components': [{'components': [{'items': [{'label': u'Abrir',
                                             'name': 'abrir',
                                             'type': 'MenuItem'},
                                            {'label': u'ReCargar',
                                             'name': 'cargar',
                                             'type': 'MenuItem'},
                                            {'label': u'Guardar',
                                             'name': 'guardar',
                                             'type': 'MenuItem'}],
                                  'label': u'Archivo',
                                  'name': 'archivo',
                                  'type': 'Menu'},
                                 {'items': [{'label': u'Estado Servidores (Dummy)',
                                             'name': 'dummy',
                                             'type': 'MenuItem'},
                                            {'label': u'\xdalt. Cbte.',
                                             'name': 'lastCBTE',
                                             'type': 'MenuItem'},
                                            {'label': u'\xdalt. ID',
                                             'name': 'lastID',
                                             'type': 'MenuItem'},
                                            {'label': u'Recuperar CAE',
                                             'name': 'getCAE',
                                             'type': 'MenuItem'}],
                                  'label': u'Consultas',
                                  'name': 'consultas',
                                  'type': 'Menu'},
                                 {'items': [{'label': u'Instructivo',
                                             'name': 'instructivo',
                                             'type': 'MenuItem'},
                                            {'label': u'Acerca de',
                                             'name': 'acercade',
                                             'type': 'MenuItem'},
                                            {'label': u'Limpiar estado',
                                             'name': 'limpiar',
                                             'type': 'MenuItem'},
                                            {'label': u'Mensajes XML',
                                             'name': 'mensajesXML',
                                             'type': 'MenuItem'},
                                            {'label': u'Ver/Ocultar Estado',
                                             'name': 'estado',
                                             'type': 'MenuItem'},
                                            {'label': u'Ver Configuraci\xf3n',
                                             'name': 'configuracion',
                                             'type': 'MenuItem'}],
                                  'label': u'Ayuda',
                                  'name': 'ayuda',
                                  'type': 'Menu'}],
                  'name': 'menu',
                  'type': 'MenuBar'},
                 {'bgcolor': u'#F9F9F8',
                  'components': [{'left': '18',
                                  'name': 'lblWebservice',
                                  'text': u'Webservice:',
                                  'top': '10',
                                  'type': 'Label'},
                                 {'items': [u'wsfe',
                                            u'wsfev1',
                                            u'wsfexv1'],
                                  'left': '102',
                                  'name': 'cboWebservice',
                                  'selection': 0,
                                  'text': u'wsfe',
                                  'top': '4',
                                  'type': 'ComboBox',
                                  'value': u'wsfe',
                                  'width': '89'},
                                 {'label': u'Marcar Todo',
                                  'left': '313',
                                  'name': 'btnMarcarTodo',
                                  'tooltip': u'Seleccionar todas las facturas',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'label': u'Autorizar Lote',
                                  'left': '199',
                                  'name': 'btnAutorizarLote',
                                  'tooltip': u'Obtener CAE para todas las facturas',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'label': u'Previsualizar',
                                  'left': '414',
                                  'name': 'btnPrevisualizar',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'label': u'Autenticar',
                                  'left': '19',
                                  'name': 'btnAutenticar',
                                  'tooltip': u'Iniciar Sesin en la AFIP',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'font': {'size': 8, 'family': 'sans serif', 'face': u'Sans'},
                                  'height': '190',
                                  'left': '18',
                                  'multiline': True,
                                  'name': 'txtEstado',
                                  'text': u'\n',
                                  'top': '300',
                                  'type': 'TextBox',
                                  'value': u'\n',
                                  'width': '554'},
                                 {'left': '20',
                                  'name': 'lblProgreso',
                                  'text': u'Progreso:',
                                  'top': '194',
                                  'type': 'Label'},
                                 {'left': '23',
                                  'name': 'lblEstado',
                                  'text': u'Estado:',
                                  'top': '280',
                                  'type': 'Label'},
                                 {'label': u'Enviar',
                                  'left': '514',
                                  'name': 'btnEnviar',
                                  'tooltip': u'Generar y enviar mails',
                                  'top': '163',
                                  'type': 'Button',
                                  'width': '53'},
                                 {'editable': False,
                                  'left': '260',
                                  'name': 'txtArchivo',
                                  'text': u'facturas.csv',
                                  'top': '5',
                                  'type': 'TextBox',
                                  'value': u'facturas.csv',
                                  'width': '313'},
                                 {'left': '195',
                                  'name': 'lblArchivo',
                                  'text': u'Archivo:',
                                  'top': '10',
                                  'type': 'Label'},
                                 {'label': u'Autorizar',
                                  'left': '109',
                                  'name': 'btnAutorizar',
                                  'tooltip': u'Obtener CAE por cada factura',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'bgcolor': u'#FFFFFF',
                                  'font': {'size': 8, 'face': u'Tahoma'},
                                  'height': '106',
                                  'item_count': 0,
                                  'left': '18',
                                  'name': 'lvwListado',
                                  'sort_column': -1,
                                  'top': '53',
                                  'type': 'ListView',
                                  'width': '556'},
                                 {'left': '18',
                                  'name': 'lblFacturas',
                                  'text': u'Facturas:',
                                  'top': '35',
                                  'type': 'Label',
                                  'width': '117'},
                                 {'bgcolor': u'#D1C2B6',
                                  'height': '16',
                                  'left': '113',
                                  'name': 'pbProgreso',
                                  'top': '195',
                                  'type': 'Gauge',
                                  'width': '453'},
                                 {'filename': u'logo-sistemasagiles.png',
                                  'height': '40',
                                  'left': '23',
                                  'name': 'image_63_220',
                                  'onmousedclick': u'import webbrowser; webbrowser.open_new("http://www.sistemasagiles.com.ar/")',
                                  'top': '224',
                                  'type': 'Image',
                                  'width': '142'},
                                 {'filename': u'logo-pyafipws.png',
                                  'left': '469',
                                  'name': 'image_432_229',
                                  'onmousedclick': u'import webbrowser; webbrowser.open_new("http://www.pyafipws.com.ar/")',
                                  'top': '212',
                                  'type': 'Image'}],
                  'fgcolor': u'#4C4C4C',
                  'height': '600',
                  'id': 196,
                  'image': '',
                  'label': u'',
                  'left': '1',
                  'name': u'panel',
                  'top': '0',
                  'type': 'Panel',
                  'width': '592'}],
  'height': '300px',
  'image': '',
  'name': 'bgTemplate',
  'title': u'Aplicativo Factura Electr\xf3nica (PyRece)',
  'type': 'Window',
  'width': '592px'}]
