from setuptools import setup

setup(
    name='py3afipws',
    packages=['py3afipws'],  # this must be the same as the name above
    version='0.10.26',
    description='pyafipws for python3',
    long_description='Factura Electrónica AFIP y otros servicios web',
    author='Mastercore',
    author_email='far@mastercore.net',
    # use the URL to the github repo
    url='https://github.com/odoo-mastercore/pyafipws',
    download_ur3='https://github.com/odoo-mastercore/py3afipws/tarball/0.10.11',
    keywords=['py3afipws', 'py3'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)
